class Contract < ActiveRecord::Base
  include Redmine::SafeAttributes

  belongs_to :user
  belongs_to :project
  belongs_to :counter_party, :class_name => "Entity"
  belongs_to :created_by, :class_name => "User"
  belongs_to :updated_by, :class_name => "User"
  acts_as_customizable
  acts_as_watchable

  validates_presence_of :contract_identifier, :user, :counter_party, :date_valid_from, :date_valid_to, :reward_unit, :unit_price
  validates_uniqueness_of :contract_identifier, :if => Proc.new { |contract|
                          contract.contract_identifier_changed? && contract.contract_identifier.present?
                        }, :case_sensitive => false
  validate :validate_contract

  scope :visible, lambda { |*args|
    where(Contract.visible_condition(args.shift || User.current, *args))
  }
  # Contract Types
  CONTRACT_TYPE_OPTIONS = [
    ["proper", :label_contract_type_option_proper],
    ["dispatch", :label_contract_type_option_dispatch],
    ["quasi_delegation", :label_contract_type_option_quasi_delegation],
    ["deliverable", :label_contract_type_option_deliverable],
  ]

  # Reward Units
  REWARD_UNIT_OPTIONS = [
    ["hour", :label_reward_unit_option_hour],
    ["day", :label_reward_unit_option_day],
    ["week", :label_reward_unit_option_week],
    ["month", :label_reward_unit_option_month],
    ["year", :label_reward_unit_option_year],
    ["deliverable", :label_reward_unit_option_deliverable],
  ]

  scope :contract_type, lambda { |arg| where(arg.blank? ? nil : { :contract_type => arg }) }

  scope :sorted, lambda { order(*Contract.fields_for_order_statement) }

  scope :like, lambda { |q|
    q = q.to_s
    if q.blank?
      where({})
    else
      pattern = "%#{q}%"
      sql = +"LOWER(#{table_name}.contract_identifier) LIKE LOWER(:p)"
      params = { :p => pattern }

      where(sql, params)
    end
  }

  # Returns a SQL conditions string used to find all contracts visible by the specified user
  def self.visible_condition(user, options = {})
    sql = "1=0"
    sql = "1=1" if user.allowed_to?(:view_contracts, nil, :global => true) || user.admin?
    sql
  end

  safe_attributes(
    "user_id",
    "project_id",
    "contract_identifier",
    "contract_type",
    "counter_party_id",
    "date_valid_from",
    "date_valid_to",
    "reward_unit",
    "unit_price",
    "original_currency",
    "original_unit_price",
    "exchange_rate",
  )

  def visible?(user = User.current)
    user.allowed_to?(:view_contracts, nil) || user.admin?
  end

  def editable?(user = User.current)
    user.allowed_to?(:edit_contracts, nil) || user.admin?
  end

  def deletable?
    user.allowed_to?(:edit_contracts, nil) || user.admin?
  end

  # inspired by User#valid_notification_options
  def valid_contract_type_options
    self.class.valid_contract_type_options(self)
  end

  def valid_reward_unit_options
    self.class.valid_reward_unit_options(self)
  end

  # inspired by User#self.valid_notification_options(user=nil)
  def self.valid_contract_type_options(contract = nil)
    if contract.nil?
      CONTRACT_TYPE_OPTIONS.reject { |option| option.first == "selected" }
    else
      CONTRACT_TYPE_OPTIONS
    end
  end

  def self.valid_reward_unit_options(contract = nil)
    # Note that @user.membership.size would fail since AR ignores
    # :include association option when doing a count
    if contract.nil?
      REWARD_UNIT_OPTIONS.reject { |option| option.first == "selected" }
    else
      REWARD_UNIT_OPTIONS
    end
  end

  def self.create_user_contract(user, attributes)
    contract = []
    if attributes
      project_ids = Array.wrap(attributes[:project_ids] || attributes[:project_id])
      role_ids = Array.wrap(attributes[:role_ids])
      project_ids.each do |project_id|
        member = Member.find_or_new(project_id, principal)
        member.role_ids |= role_ids
        member.save
        members << member
      end
    end
    members
  end

  def validate_contract
    if date_valid_from && date_valid_to && (date_valid_from_changed? || date_valid_to_changed?) && date_valid_to < date_valid_from
      errors.add :date_valid_to, :greater_than_date_valid_from
    end

    if contract_type == "quasi_delegation" && (!user || !project)
      errors.add :contract_type, :requires_user_and_project
    end

    if contract_type == "dispatch" || contract_type == "proper"
      if !user
        errors.add :contract_type, :requires_user
      else
        overlap = false
        if c = user.valid_contract
          if range = c.date_valid_from..c.date_valid_to
            overlap = true if range.cover?(date_valid_from) || range.cover?(date_valid_to)
          end
        end
        if overlap
          errors.add :date_valid_from, :overlap_existing_valid_contract
        end
      end
    end
  end

  def is_valid?
    date_valid_from && date_valid_to && Time.now <= date_valid_to
  end

  def is_expired?
    !is_valid?
  end

  def is_current?
    date_valid_from && date_valid_to && date_valid_from <= Time.now && Time.now <= date_valid_to
  end

  def self.find(*args)
    if args.first && args.first.is_a?(String) && !/^\d*$/.match?(args.first)
      contract = find_by_contract_identifier(*args)
      raise ActiveRecord::RecordNotFound, "Couldn't find contract with identifier=#{args.first}" if contract.nil?
      contract
    else
      super
    end
  end

  # Returns a string of css classes that apply to the contract
  def css_classes(user = User.current)
    s = +"contract"
    s << " expired" if is_expired?
    s << " valid" if is_valid?
    s << " current" if is_current?
    s
  end
end
